// Persistence Of Vision Ray Tracer Scene Description File
// -------------------------------------------------------
// File: @(#)glowlits.pov
// Description: Here the glow effects are coupled with light sources.
// Features demonstrated: glow
// Creation Date: $ 27 Sep 2000, 16:08:45 $
// Last modified: $ 14 Jun 2004, 22:36:41 $
// Author: Chris Huff
//
// -w320 -h240 +a0.3

//#version unofficial megapov 1.1;
#version 3.7;                   

global_settings {
	assumed_gamma 1.0
}

#declare cameraLocation = <-5, 6, -18>;
#declare cameraLookAt = <0, 1.5, 0>;

camera {
	location cameraLocation
	up y*image_height right x*image_width
	look_at cameraLookAt
	angle 45
}

#macro PrintVector(v0)
    #debug concat("<", vstr(3, v0, ", ", 0, -1), "> \n")
#end

// ----------------------------------------
//  Macros and constants from PovRay Source   
// ----------------------------------------
#local M_PI_2 =  1.57079632679489661923;
// ----------------------------------------
 
//  Vector subtraction macro from povray source
#macro VSub(outputVector, inputVector0, inputVector1)
	#declare outputVector = <inputVector0.x - inputVector1.x, inputVector0.y - inputVector1.y, inputVector0.z - inputVector1.z>;
#end

//  Dot Product macro from povray source
#macro VDot(ScalarAngle, inputVector0, inputVector1)
	#declare ScalarAngle = ((inputVector0.x * inputVector1.x) + (inputVector0.y * inputVector1.y) + (inputVector0.z * inputVector1.z));
#end

//  Vector Scale macro from povray source
#macro VScale(outputVector, inputVector, ScaleFactor)
	#declare outputVector = <inputVector.x * ScaleFactor, inputVector.y * ScaleFactor, inputVector.z * ScaleFactor>;
#end

//  Vector Addition macro from povray source
#macro VAddEq(outputVector, inputVector)
	#declare outputVector = <outputVector.x + inputVector.x, outputVector.y + inputVector.y, outputVector.z + inputVector.z>;
#end

//  Vector Subtraction macro from povray source
#macro VSubEq(outputVector, inputVector)
	#declare outputVector = <outputVector.x - inputVector.x, outputVector.y - inputVector.y, outputVector.z - inputVector.z>;
#end

// Vector Length - returs Scalar Euclidean Length (a) of Vector (b)
#macro VLength(vectorLength, inputVector)
    #declare vectorLength = sqrt((inputVector.x * inputVector.x) + (inputVector.y * inputVector.y) + (inputVector.z * inputVector.z));
#end

//  Scale Color macro from PovRay source
#macro Scale_Color(outputColor, inputColor, factor)
  //#declare outputColor = <outputColor.x * factor, outputColor.y * factor, outputColor.z * factor, outputColor.filter * factor, outputColor.transmit * factor>;
  #declare outputColor = <outputColor.x * factor, outputColor.y * factor, outputColor.z * factor> 
#end

//  Add Color macro from PovRay source
#macro Add_Color(outputColor, inputColor0, inputColor1)
  //#declare outputColor = <inputColor0.x + inputColor1.x, inputColor0.y + inputColor1.y, inputColor0.z + inputColor1.z, inputColor0.filter + inputColor1.filter, inputColor0.transmit + inputColor1.transmit>;  
    #declare outputColor = <inputColor0.x + inputColor1.x, inputColor0.y + inputColor1.y, inputColor0.z + inputColor1.z> 
#end

// ----------------------------------------

#macro CalculateGlowFalloff(fade_pow, Glow_Type, Center, Cutoff_Radius, Size, Color, Depth, Max_Distance, RayDirection, RayInitial)
	#local lightDir = <0, 0, 0>;                                        /*direction of glow source*/
	VSub(lightDir, RayInitial, Center)
	
	#local scattered_light = <0, 0, 0>; // color
	#local Scattering = 0;
	#local cosA = 0;                                                    /*cosine of the angle between the ray and the direction of the glow source*/
	VDot(cosA, lightDir, RayDirection)
	
	#local Dist = 0;
	
    #local Pt = <0, 0, 0>;                                              /*Point on plane perpendicular to ray and passing through glow*/   
    VScale(Pt, RayDirection, -cosA)	
	VAddEq(Pt, RayInitial);           		
	VSubEq(Pt, Center);
	VLength(Dist, Pt);
	
	#if(Cutoff_Radius = 0 | Dist < Cutoff_Radius)	
		/* scattered energy integration along ray */
		#switch(Glow_Type)		
			#case 0:/* A model, I(d) = 1/d^2 */
				#local Dist = Dist / Size;
				#local Scattering = (atan((Depth + cosA) / Dist) - atan(cosA / Dist)) / Dist;
			#break

			#case 1:/* B model, I(d) = 1/(d^2+1) */			
				#local d0 = <0, 0, 0>;
				VDot(d0, Pt, Pt);
				
				#local denom = sqrt(d0 + 1) / Size;
				
				#if(Depth >= Max_Distance) /* Optimization */
					#local Scattering = (M_PI_2 - atan(cosA / denom)) / denom;
				#else
					#local Scattering = (atan((Depth + cosA) / denom) - atan(cosA / denom)) / denom;
				#end			
		    #break

			#case 2:/*exp() falloff*/
				#if(cosA < 0)
					#local Scattering = exp(-Dist / Size); 
				#end
			#break;

			#case 3:/*Cosine falloff*/
				#local d = Dist / Size;
				#if(d < 1 & cosA < 0)
					#local Scattering = sin(max(0.0, 1 - d) * M_PI_2);
				#end
			#break
		#end
		
		#if(fade_pow != 1)
			#if(fade_pow = 2)
				#local Scattering = Scattering * Scattering;
			#else
				#local Scattering = pow(Scattering, fade_pow);
		    #end
		#end
		
	    Scale_Color(scattered_light, Color, Scattering);
        Add_Color(Color, Color, scattered_light);
	#end
#end
   
#macro BlackHoleWarp(Iterations, Falloff, Strength, Position, Radius)
    #local counter = 0; // set counter to 0
    #while (counter < Iterations) 
   
        warp 
        {
            black_hole Position, Radius
            falloff Falloff
            strength Strength
            turbulence 0.4
        }
   
        #local counter = counter + 1;
    #end
#end

#macro GlowType1Falloff(cameraLocation, cameraLookAt, glowLocation, glowSize, glowFadePower, glowColor) 
#local Depth = 5;
#local Max_Distance = 5;

//  Translate Macro Params to GlowPatch variable names, for easier debugging and comparison
#local RayInitial = cameraLocation;
#local RayDirection = cameraLookAt;

#local Center = glowLocation;
#local Size = glowSize;
#local fade_pow = glowFadePower;
#local scattered_light = glowColor;
#local Color = glowColor;

//  Calculate internal variables
#local lightDir = <0, 0, 0>;                                        /*direction of glow source*/
VSub(lightDir, RayInitial, Center)

#local cosA = 0;                                                    /*cosine of the angle between the ray and the direction of the glow source*/
VDot(cosA, lightDir, RayDirection)

#local Pt = <0, 0, 0>;                                              /*Point on plane perpendicular to ray and passing through glow*/   
VScale(Pt, RayDirection, -cosA)	
VAddEq(Pt, RayInitial);           		
VSubEq(Pt, Center); 

#local d0 = <0, 0, 0>;
VDot(d0, Pt, Pt);
				
#local denom = sqrt(d0 + 1) / Size;
				
#if(Depth >= Max_Distance) /* Optimization */
    #local Scattering = (M_PI_2 - atan(cosA / denom)) / denom;
#else
	#local Scattering = (atan((Depth + cosA) / denom) - atan(cosA / denom)) / denom;
#end
				
#if(fade_pow != 1)
    #if(fade_pow = 2)
	    #local Scattering = Scattering * Scattering;
	#else
		#local Scattering = pow(Scattering, fade_pow);
	#end
#end
		
Scale_Color(scattered_light, Color, Scattering);
Add_Color(Color, Color, scattered_light);


//#debug concat("[", str(numElements,0,0), "] \n")
//#debug concat("[", str(numElementsTotal,0,0), "] \n")
//#local fname = "foo.inc";
//#fopen filehandle fname write
//#write (filehandle, indexToUse, "\n")                       
//#fclose filehandle 
Color
#end				

#macro CalculatePercent(Value, Total)
    ((Value / Total) * 100)
#end

#macro GlowInterior(FuzzCount, glowRadius)
    #local j = FuzzCount;
    #while (j > 0)
    density
    {
        spherical
        //scale j * 0.9
        scale j * 0.125
        density_map
        {                         
            [0.00 rgb glowColor * 00.000]/*
            [0.10 rgb glowColor * 00.050]
            [0.20 rgb glowColor * 00.100]
            [0.80 rgb glowColor * 00.250]
            [0.90 rgb glowColor * 00.750]
            [0.95 rgb glowColor * 04.000]
            [1.00 rgb glowColor * 10.000]*/
            #local k = 1;
            #local kMax = 255;
            #while (k < kMax)
                
                #local percentOfTotal = CalculatePercent(k, kMax);
                    
                #local factor = pow(log(exp(k)), ln(percentOfTotal)) * 0.0001;
                [k * 0.01 rgb (glowColor * (factor))]

            #local k = k + 1;
            #end
        }
    }
    #local j = j - 1;
    #end
#end


#macro Glow(glowPosition, glowRadius, glowColor)
#local coreScaleFactor = 0.030625;   //0.00765625//0.0153125//0.030625 

#local lightFactor = (1 / 24);
#local density0 = 1;
#local density1 = 0;
#local density2 = 1;

#local enableSurroundingFog = true;
#local enablePrimaryHalo = true;  // get "FixedSimpleVector internal limit exceeded" error when this = true and coreFuzzCount > 17
#local enableCoreFuzz = 0;     // can also produce this error with various combinations of fuzzcount, primary halo enabled, and then rendering a selection
#local enableCoreFuzz0 = 0;
#local enableCoreFuzz1 = 0;
#local coreFuzzCount = 17;

#if (enableSurroundingFog)
//  Surrounding "Fog"
light_source 
{
    glowPosition       
    color rgb glowColor * lightFactor
    fade_distance 4 
    fade_power 2
    looks_like 
    { 
        sphere 
        { 
            <0, 0, 0>, glowRadius * 50
            hollow 
            pigment 
            { 
                rgbt <0, 0, 0, 1> 
            }
            interior 
            {
                media 
                {
                    intervals 2         
                    samples 10, 100        
                    emission glowColor
                    method 2
                    density
                    {
                        spherical
                        scale 8
                        density_map
                        {                         
                            [0.0 rgb glowColor * 0.0000]
                            [0.8 rgb glowColor * 0.00765625]
                        }                        
                    }
                }
            }
            finish
            {
                reflection 0
                diffuse 0
                ambient 0
                phong 0
                specular 0
                brilliance 0                
            }             
        } 
    } 
}
#end

#if (enablePrimaryHalo)
//  Primary "Halo"
light_source 
{
    glowPosition       
    color rgb glowColor * lightFactor
    fade_distance 4 
    fade_power 2
    looks_like 
    {
        sphere 
        { 
            <0, 0, 0>, glowRadius * 6
            hollow 
            pigment 
            { 
                rgbt <0, 0, 0, 1> 
            }
            interior 
            {
                
                media 
                {
                    //intervals 2
                    samples 250, 500
                    emission glowColor
                    method 2
                    GlowInterior(2, glowRadius)
                }
                /**/
                /*
                media 
                {
                    intervals 2
                    samples 250, 500
                    emission glowColor
                    method 2
                    density
                    {
                        spherical
                        scale scaleFactor
                        density_map
                        {
                            [0.0 rgb glowColor * 0.0000]

                            #local j = 1;
                            #local jMax = 255;
                            #while (j < jMax)

                                #local percentOfTotal = CalculatePercent(j, jMax);

                                #local factor = pow(log(exp(j)), ln(percentOfTotal)) * 0.0001;
                                [j * 0.01 rgb (glowColor * (factor))]

                            #local j = j + 1;
                            #end
                        }
                    }
                }
                */
                
                media 
                {
                    intervals 2
                    samples 20, 100
                    emission glowColor
                    method 2
                    #if (density0)
                        density
                        {
                            spherical
                            scale 0.75
                            density_map
                            {
                                [0.0 rgb glowColor * 0.000]
                                //[0.8 rgb glowColor * 0.250]
                                //[0.9 rgb glowColor * 0.750]
                                [1.0 rgb glowColor * 1.000]
                            }
                        }
                    #end
                    #if (density1)
                        density
                        { 
                            spherical
                            density_map
                            {
                                [0.00 color rgb glowColor * 0.000]
                                [0.90
                                    granite
                                    color_map 
                                    {
                                        [0 color rgb glowColor * 1.750]
                                        [1 color rgb glowColor * 1.250]
                                    }
                                    //  BlackHoleWarp(Iterations, Falloff, Strength, Position, Radius)
                                    BlackHoleWarp(4, 1, 1, <0, 0, 0>, 1)
                                ]
                                [1.00 color rgb glowColor * 1.2500]
                            }
                        } 
                    #end 
                    #if (density2)
                        density
                        { 
                            spherical
                            density_map
                            {
                                [0.00 color rgb glowColor * 0.000]
                                [ 0.90
                                    granite
                                    color_map 
                                    {
                                        [0 color rgb glowColor * 2.750]
                                        [1 color rgb glowColor * 3.250]
                                    }
                                ]
                                [1.00 color rgb glowColor * 3.500]
                            }
                        } 
                    #end
                }
                /**/
            }            
            finish
            {
                reflection 0
                diffuse 0
                ambient 0
                phong 0
                specular 0
                brilliance 0                
            }             
        } 
    } 
}
/*
light_source 
{
    glowPosition
    color rgb glowColor * lightFactor
    fade_distance 4 
    fade_power 2
    looks_like 
    { 
        sphere 
        { 
            <0, 0, 0>, glowRadius
            hollow 
            pigment 
            { 
                rgbt <0, 0, 0, 1> 
            }
            interior 
            {
                media 
                {
                    intervals 2
                    samples 20, 100
                    emission glowColor
                    method 2
                    GlowInterior(4)
                }
            }
            finish
            {
                reflection 0
                diffuse 0
                ambient 0
                phong 0
                specular 0
                brilliance 0
            }
        } 
    } 
}
*/
#end

#if (enableCoreFuzz0)
//  Core "Fuzz"
#local j = 1;
//#local j = coreFuzzCount;
//#while (j > 0)
    light_source 
    {
        glowPosition       
        color rgb (glowColor * lightFactor) * 2
        fade_distance 4 
        fade_power 2
        looks_like 
        {       
            sphere 
            { 
                //<0, 0, 0>, (glowRadius / (j + 0.01))
                //<0, 0, 0>, ((glowRadius * 0.25) / (j * 0.05))
                <0, 0, 0>, (glowRadius / ((j * j) * 0.01))
                hollow 
                pigment 
                { 
                    rgbt <0, 0, 0, 1> 
                }
                interior 
                {
                    media 
                    {
                        intervals 2         
                        samples 10, 100          
                        emission glowColor * 10
                        method 2
                        #if (density0) 
                            density
                            {
                                spherical
                                scale 0.06125 // comment to make huge balls
                                density_map
                                {
                                    [0.0 rgb glowColor * 0.0000]  
                                    #local j = 1;
                                    #while (j < 255) 
                                        #local factor = 0;
                                        #local tempGlowColor = glowColor;

                                        //#if (j < 75)                                 
                                        // #local factor = j / 100;
                                        //#else
                                        //#local factor = log(j) + sin((j)*(3.14*2));//(j / 5) + log(j);
                                        //#local factor = sin(max(0.0, 1 - d) * M_PI_2)
                                        //#end
                                        
                                        #local Color = GlowType1Falloff(cameraLocation, cameraLookAt, glowPosition, (glowRadius / ((j * j) * 0.01)), 0.5, tempGlowColor);
                                        [j * 0.0255 rgb Color * 5]

                                        //[j * 0.01 rgb glowColor * (factor)]
                                    #local j = j + 1;
                                    #end
                                    //[1 rgb glowColor * 30]                                    
                                }                        
                            }
                        #end
                    }
                }
                finish
                {
                    //emission 1
                    reflection 0
                    diffuse 0
                    ambient 0
                    phong 0
                    specular 0
                    brilliance 0                
                }             
            } 
        } 
    }
    //#local j = j - 1;
    //#end
     
#end

#if (enableCoreFuzz1)
//  Core "Fuzz"
#local j = 1;
#while (j > 0)

    light_source 
    {
        glowPosition       
        color rgb (glowColor * lightFactor) / 30
        fade_distance 4 
        fade_power 2
        looks_like 
        {       
            sphere 
            { 
                //<0, 0, 0>, (glowRadius / (j + 0.01))
                <0, 0, 0>, ((glowRadius * 0.0006125) / (j * 0.1))
                hollow 
                pigment 
                { 
                    rgbt <0, 0, 0, 1> 
                }
                interior 
                {
                    media 
                    {
                        intervals 2         
                        samples 10, 100          
                        emission glowColor
                        method 2
                        scale 0.1
                        #if (density0) 
                            density
                            {
                                spherical
                                density_map
                                {
                                    [0.0 rgb glowColor * 0.0000]
                                    [0.8 rgb glowColor * j]
                                }                        
                            }
                        #end
                    }
                }
                finish
                {
                    reflection 0
                    diffuse 0
                    ambient 0
                    phong 0
                    specular 0
                    brilliance 0                
                }             
            } 
        } 
    }
    #local j = j - 1;
    #end
     
#end

#end   
// ---------------------------------------- 
  
   
//Types:
#local Float = 0;
#local Vector = < 0, 0, 0>;
#local Color = color rgb < 0, 0, 0>;

#macro SRand(RS) (rand(RS)*2 - 1) #end
#macro RRand(RS, Min, Max) (rand(RS)*(Max-Min) + Min) #end
#macro Clamp(V, Min, Max) (min(Max, max(Min, V))) #end
#macro Range(V, Rmn, Rmx) (V*(Rmx-Rmn) + Rmn) #end
#macro RClamp(V, Rmn, Rmx, Min, Max) (Clamp(Range(V, Rmn, Rmx), Min, Max)) #end

#local RsA = seed(574647);

#local J = Float;
#local J = 0;
#while(J<52)
    Glow(< SRand(RsA)*8, rand(RsA)*4, SRand(RsA)*8>, rand(RsA)*1.5, 0.65*< RRand(RsA, 0.1, 1), RRand(RsA, 0.1, 1), RRand(RsA, 0.1, 1)>)
	#local J = J + 1;
#end

box {
	<-15, 0,-10>, < 15, 0, 10>
	pigment {checker color rgb 0.05, color rgb 1}
	finish {
		reflection {0.5, 0.75 metallic }
		diffuse 0.25 ambient 0
	}
}

sphere {
	< 0, 2, 0>, 2
	pigment {rgb 1}
	finish {
		reflection { 0.5, 0.85 metallic }
		diffuse 0.15 ambient 0
	}
}
